/*
Copyright (c) 2003 by Radim HSN Kolar (hsn@cybermail.net)
 
You may copy or modify this file in any manner you wish, provided
that this notice is always included, and that you hold the author
harmless for any loss or damage resulting from the installation or
use of this software.
 
		     This is a free software.  Be creative. 
		    Let me know of any bugs and suggestions.
 
Version 1.0rc3

---------------------

Modified from the original by Mark Petrovic, EarthLink, Inc.
  - added Log4J logging statements, April, 2004
  
*/
package net.earthlink.research.fsp;

import java.net.InetAddress;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.SocketException;
import java.net.SocketTimeoutException;

import java.io.InterruptedIOException;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.apache.log4j.ConsoleAppender;

/** This class represents one live FSP session.
 * <p>
 * FSPsession class takes care about FSP session management. It handles
 * packet resends on timeouts and key management.
 *
 * @author Radim Kolar
 * @see FSPpacket
 * @version 1.0
 */
public class FSPsession {

   private Logger          logger;
   private PatternLayout   layout;
   private ConsoleAppender appender;

   private DatagramSocket socket;
   private DatagramPacket udp;
   private DatagramPacket receiveUDP;

   //private FSPpacket   packet;
   private FSPpacket   sendPacket;
   private FSPpacket   receivePacket;

   private short seq;  /* sequence number */

   private int timeout;
   private int delay;
   private int maxdelay;

   private int port;
   private InetAddress host;
   private String hostadr;  /* 1.2.3.4:2234  */

   private static Hashtable locks;

   /** minimum resend delay */
   public final static int MIN_DELAY=1000;

   /** default resend delay */
   public final static int DEFAULT_DELAY=1340;
   //public final static int DEFAULT_DELAY=5000;

   /** maximum resend delay */
   public final static int MAX_DELAY=300000;

   /** default timeout (msec) */
   public final static int DEFAULT_TIMEOUT=180000;

   /** Creates a new FSP session
    *
    * @param host hostname of fsp server
    * @param port port number on fsp server
    * @since 1.0
    */
   public FSPsession(String host, int port) throws java.net.SocketException,java.net.UnknownHostException {
      this(InetAddress.getByName(host),port);
   }

   /** Creates a new FSP session
    *
    * @param host address of fsp server
    * @param port port number on fsp server
    * @since 1.0
    */
   public FSPsession(InetAddress host, int port) throws java.net.SocketException {
      socket=new DatagramSocket();

      udp        = new DatagramPacket(new byte[FSPpacket.MAXSIZE],FSPpacket.MAXSIZE,host,port);
      //receiveUDP = new DatagramPacket(new byte[FSPpacket.MAXSIZE],FSPpacket.MAXSIZE);
      receiveUDP = new DatagramPacket(new byte[FSPpacket.MAXSIZE],FSPpacket.MAXSIZE, host, port);

      if(port==0)
         port=21;
      udp.setPort(port);
      udp.setAddress(host);
      //packet=new FSPpacket();
      sendPacket=new FSPpacket();

      // msp
      receivePacket=new FSPpacket();

      seq=(short)((int)(Math.random()*0xffff) & 0xff00);
      timeout=DEFAULT_TIMEOUT;
      delay=DEFAULT_DELAY;
      maxdelay=MAX_DELAY;

      if(locks==null)
         locks=new Hashtable();
      hostadr=host.getHostAddress()+":"+port;
      synchronized(locks) {
         Short k=(Short)locks.get(hostadr);
         if(k==null) {
            k=new Short((short)0);
            locks.put(hostadr,k);
         }
         /* Get object from hashtable, we need it for correct locking */
         java.util.Enumeration en=locks.keys();
         while(en.hasMoreElements()) {
            String hkey;
            hkey=(String)en.nextElement();
            if(hostadr.equals(hkey)) {
               hostadr=hkey;
               break;
            }
         }
      }
      this.port=port;
      this.host=host;

      logger   = Logger.getLogger(this.getClass().getPackage().getName());
      
      // Somewhat lame protection against adding redundant appenders...  msp 5May2004
      
      Enumeration e = logger.getAllAppenders();
      if( ! e.hasMoreElements() ) {
         layout   = new PatternLayout("%d{dd MMM yyyy HH:mm:ss}, %l, %m %n");
         appender = new ConsoleAppender(layout);
         logger.addAppender(appender);
         logger.setLevel(Level.WARN);
      }
   }

   /** Sends FSP packet and waits for reply, resend packet if needed
    *
    * @since 1.0
    * @throws java.io.InterruptedIOException
    */

   public FSPpacket interact(byte cmd,long filepos, byte data1[],int offset1,int length1, byte data2[],int offset2,int length2) throws InterruptedIOException {

      int     rdelay;
      int     rtimeout;

      rdelay   = delay;
      rtimeout = 0;

      sendPacket.cmd    = cmd;
      seq               += 0x0100;
      sendPacket.bb_pos = (int )(filepos & 0xffffffff);
      sendPacket.bb_seq = seq;

      sendPacket.setData(data1, offset1, length1, data2, offset2, length2);
      
      sendPacket.bb_seq = (short )(seq + (++sendPacket.bb_seq & 0xff));

      synchronized(hostadr) {

         sendPacket.bb_key = ( (Short )locks.get(hostadr) ).shortValue();
         sendPacket.assemble(udp);

         setSocketTimeout(socket, rdelay);
         transmitPacket(udp);
         while(true) {
            try {
               socket.receive(receiveUDP);
               if( receivePacket.disassemble(receiveUDP) ) {
                  if ( receivePacket.bb_seq == sendPacket.bb_seq ) {
                     locks.put(hostadr, new Short(receivePacket.bb_key));
                     return receivePacket;
                  }
               }
            } 
            catch (SocketTimeoutException e) {
               rtimeout += rdelay;
               if(rtimeout >= timeout)
                  throw new InterruptedIOException("FSP application receive timeout");
               rdelay *= 1.5f;
               if( rdelay > maxdelay )
                  rdelay = maxdelay;

               logger.warn("FSP request timeout.  Retrying...");

               setSocketTimeout(socket, rdelay);
               transmitPacket(udp);
            }
            catch (InterruptedIOException ioe) {
               ioe.printStackTrace();
            }
            catch (IOException ioe) {
               ioe.printStackTrace();
            }
         }
      }
   }

   private void setSocketTimeout(DatagramSocket sock, int timeout) {
      try {
         sock.setSoTimeout(timeout);
      }
      catch (SocketException ex) {
         ex.printStackTrace();
      }
   }

   private void transmitPacket(DatagramPacket pack) {
      try {
         socket.send(udp);
      }
      catch (IOException ex) {
         ex.printStackTrace();
      }
   }

   /** close a session.
    * <p>
    * Session object can't be used after session is closed. This also sends
    * CC_BYE command to server.
    * 
    * @since 1.0
    */
   public void close() {
      if(socket!=null)
         try {
            interact(FSPpacket.CC_BYE,0,null,0,0,null,0,0);
            socket.close();
         } catch (Exception e) {}

      socket=null;
      udp=null;
      
      receiveUDP=null;
      //packet=null;
      sendPacket=null;
      receivePacket=null;
   }

   /** gets the delay.
    * @since 1.0
    */
   public int getDelay() {
      return delay;
   }
   /** sets the delay.
    * <p>
    * This functions sets delay parameter of FSP protocol stack. This
    * is timeout value for first packet. If packet is lost delay is multiplied
    * by 1.5. Delay is in milliseconds and must be between MIN_DELAY and MAX_DELAY
    *
    * @since 1.0
    * @param ndelay new delay in milliseconds
    */
   public void setDelay(int ndelay) {
      if(ndelay<MIN_DELAY)
         ndelay=MIN_DELAY;
      else
         if(ndelay>MAX_DELAY)
            ndelay=MAX_DELAY;

      this.delay=ndelay;
   }
   /** gets the maximum delay between packets.
    *
    * @since 1.0
    */
   public int getMaxDelay() {
      return maxdelay;
   }
   /** sets the maximum delay between packets.
    *
    * @param newdelay new max delay time in milliseconds.
    * @since 1.0
    */
   public void setMaxDelay(int newdelay) {
      if(newdelay>MAX_DELAY)
         newdelay=MAX_DELAY;
      this.maxdelay=newdelay;
   }

   /** get timeout value.
    * @since 1.0
    */
   public int getTimeout() {
      return timeout;
   }
   /** set timeout value.
    * <p>
    * If no packet from server is received in this time, session times out.
    * 
    * @since 1.0
    * @param timeout timeout value in milliseconds
    */
   public void setTimeout(int ntimeout) {
      if(ntimeout>=0)
         this.timeout=ntimeout;
   }

   /** closes session before doing GC.
    * <p>
    * If session is not closed, close it before doing GC on this object.
    * Timeout is set to 7 sec when doing it.
    * 
    * @since 1.0
    */
   public void finalize() {
      timeout=7000;
      close();
   }

}
